# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fnmatch
from functools import cached_property
from pydantic_core import PydanticCustomError, ErrorDetails
from pydantic import (
    Field,
    computed_field,
    field_validator,
    BeforeValidator,
    ValidationError,
    ValidationInfo,
    model_validator,
)
import re
from typing import Optional, Any, Annotated, Union

from pisek.utils.text import tab
from pisek.utils.text import eprint, colored, warn
from pisek.env.base_env import BaseEnv
from pisek.config.config_hierarchy import ConfigValue, TaskConfigError, ConfigHierarchy
from pisek.config.config_types import (
    TaskType,
    OutCheck,
    JudgeType,
    DataFormat,
    Scoring,
    ProgramType,
    CMSFeedbackLevel,
    CMSScoreMode,
)
from pisek.env.context import init_context
from pisek.jobs.parts.solution_result import SUBTASK_SPEC


MaybeInt = Annotated[
    Optional[int], BeforeValidator(lambda i: (None if i == "X" else i))
]
ListStr = Annotated[list[str], BeforeValidator(lambda s: s.split())]
OptionalStr = Annotated[Optional[str], BeforeValidator(lambda s: s or None)]

MISSING_VALIDATION_CONTEXT = "Missing validation context."

ValuesDict = dict[str, Union[str, "ValuesDict", dict[Any, "ValuesDict"]]]
ConfigValuesDict = dict[
    str, Union[ConfigValue, "ConfigValuesDict", dict[Any, "ConfigValuesDict"]]
]


def _to_values(config_values_dict: ConfigValuesDict) -> ValuesDict:
    def convert(what: ConfigValue | dict) -> str | dict:
        if isinstance(what, ConfigValue):
            return what.value
        else:
            return {key: convert(val) for key, val in what.items()}

    return {key: convert(val) for key, val in config_values_dict.items()}


class TaskConfig(BaseEnv):
    """Configuration of task loaded from config file."""

    name: str
    contest_type: str
    task_type: TaskType
    scoring: Scoring

    solutions_subdir: str
    static_subdir: str
    data_subdir: str

    in_gen: str
    checker: OptionalStr
    out_check: OutCheck
    out_judge: Optional[str]
    judge_type: Optional[JudgeType]
    judge_needs_in: bool
    judge_needs_out: bool

    in_format: DataFormat
    out_format: DataFormat

    stub: OptionalStr
    headers: ListStr

    subtasks: dict[int, "SubtaskConfig"]

    solutions: dict[str, "SolutionConfig"]

    limits: "LimitsConfig"

    cms: "CMSConfig"

    checks: "ChecksConfig"

    @computed_field  # type: ignore[misc]
    @cached_property
    def total_points(self) -> int:
        return sum(sub.points for sub in self.subtasks.values())

    @computed_field  # type: ignore[misc]
    @property
    def subtasks_count(self) -> int:
        return len(self.subtasks)

    @computed_field  # type: ignore[misc]
    @cached_property
    def input_globs(self) -> list[str]:
        return sum((sub.all_globs for sub in self.subtasks.values()), start=[])

    @computed_field  # type: ignore[misc]
    @property
    def primary_solution(self) -> str:
        if len(self.solutions) == 0:
            raise RuntimeError("No solutions exist.")
        else:
            return [name for name, sol in self.solutions.items() if sol.primary][0]

    def get_solution_by_source(self, source: str) -> Optional[str]:
        sources = (name for name, sol in self.solutions.items() if sol.source == source)
        return next(sources, None)

    def __init__(self, **kwargs):
        value = {"subtask_count": max(kwargs["subtasks"]) + 1, "name": kwargs["name"]}

        with init_context(value):
            super().__init__(**kwargs)

    @staticmethod
    def load_dict(configs: ConfigHierarchy) -> ConfigValuesDict:
        GLOBAL_KEYS = [
            ("task", "name"),
            ("task", "contest_type"),
            ("task", "task_type"),
            ("task", "scoring"),
            ("task", "solutions_subdir"),
            ("task", "static_subdir"),
            ("task", "data_subdir"),
            ("tests", "in_gen"),
            ("tests", "checker"),
            ("tests", "out_check"),
            ("tests", "in_format"),
            ("tests", "out_format"),
            ("all_solutions", "stub"),
            ("all_solutions", "headers"),
        ]
        JUDGE_KEYS = [
            ("out_judge", None),
            ("judge_type", None),
            ("judge_needs_in", "0"),
            ("judge_needs_out", "1"),
        ]
        args: dict[str, Any] = {
            key: configs.get(section, key) for section, key in GLOBAL_KEYS
        }

        # Load judge specific keys
        for key, default in JUDGE_KEYS:
            if args["out_check"].value == "judge":
                args[key] = configs.get("tests", key)
            else:
                args[key] = ConfigValue(default, "_internal", "tests", key, True)

        section_names = configs.sections()

        # Load subtasks
        args["subtasks"] = subtasks = {}
        for section in sorted(section_names, key=lambda cv: cv.value):
            section_name = section.value
            if m := re.match(r"test(\d{2})", section_name):
                num = m[1]
                subtasks[int(num)] = SubtaskConfig.load_dict(
                    ConfigValue(str(int(num)), section.config, section.section, None),
                    configs,
                )

        args["solutions"] = solutions = {}
        for section in section_names:
            if m := re.match(r"solution_(.+)", section.value):
                solutions[m[1]] = SolutionConfig.load_dict(
                    ConfigValue(m[1], section.config, section.section, None), configs
                )

        args["limits"] = LimitsConfig.load_dict(configs)
        args["cms"] = CMSConfig.load_dict(configs)
        args["checks"] = ChecksConfig.load_dict(configs)

        return args

    @field_validator("contest_type", mode="after")
    @classmethod
    def validate_contest_type(cls, value: str) -> str:
        # TODO: Redo to general task types
        ALLOWED = ["kasiopea", "cms"]
        if value not in ALLOWED:
            raise PydanticCustomError(
                "contest_type_invalid",
                f"Must be one of ({', '.join(ALLOWED)})",
            )
        return value

    @model_validator(mode="after")
    def validate_model(self):
        if (
            self.task_type == TaskType.communication
            and self.out_check != OutCheck.judge
        ):
            raise PydanticCustomError(
                "communication_must_have_judge",
                "For communication task 'out_check' must be 'judge'",
                {"task_type": self.task_type, "out_check": self.out_check},
            )

        primary = [name for name, sol in self.solutions.items() if sol.primary]
        if len(primary) > 1:
            raise PydanticCustomError(
                "multiple_primary_solutions",
                "Multiple primary solutions",
                {"primary_solutions": primary},
            )
        if len(self.solutions) > 0 and len(primary) == 0:
            raise PydanticCustomError(
                "no_primary_solution",
                "No primary solution set",
                {},
            )

        for i in range(len(self.subtasks)):
            if i not in self.subtasks:
                raise PydanticCustomError(
                    "missing_subtask",
                    f"Missing section for subtask {i}",
                    {},
                )

        self._compute_predecessors()
        return self

    def _compute_predecessors(self) -> None:
        visited = set()
        computed = set()

        def compute_subtask(num: int) -> list[int]:
            subtask = self.subtasks[num]
            if num in computed:
                return subtask.all_predecessors
            elif num in visited:
                raise PydanticCustomError(
                    "cyclic_predecessor_subtasks", "Cyclic predecessor subtasks", {}
                )

            visited.add(num)
            all_predecessors = sum(
                (compute_subtask(p) for p in subtask.direct_predecessors),
                start=subtask.direct_predecessors,
            )

            def normalize_list(l):
                return list(sorted(set(l)))

            subtask.all_predecessors = normalize_list(all_predecessors)
            subtask.all_globs = normalize_list(
                sum(
                    (self.subtasks[p].in_globs for p in subtask.all_predecessors),
                    start=subtask.in_globs,
                )
            )
            computed.add(num)

            return subtask.all_predecessors

        for i in range(self.subtasks_count):
            compute_subtask(i)


class SubtaskConfig(BaseEnv):
    """Configuration of one subtask."""

    _section: str
    num: int
    name: str
    points: int = Field(ge=0)
    in_globs: ListStr
    all_globs: list[str] = []
    direct_predecessors: list[int]
    all_predecessors: list[int] = []

    def in_subtask(self, filename: str) -> bool:
        return any(fnmatch.fnmatch(filename, g) for g in self.all_globs)

    def new_in_subtask(self, filename: str) -> bool:
        return any(fnmatch.fnmatch(filename, g) for g in self.in_globs)

    @staticmethod
    def load_dict(number: ConfigValue, configs: ConfigHierarchy) -> ConfigValuesDict:
        KEYS = ["name", "points", "in_globs", "predecessors"]
        num = int(number.value)
        args: dict[str, Any]
        if num == 0:
            args = {key: configs.get(number.section, key) for key in KEYS}
        else:
            args = {
                key: configs.get_from_candidates(
                    [(number.section, key), ("all_tests", key)]
                )
                for key in KEYS
            }
        args["direct_predecessors"] = args.pop("predecessors")

        return {"_section": configs.get(number.section, None), "num": number, **args}

    @field_validator("in_globs", mode="after")
    @classmethod
    def validate_globs(cls, value: list[str], info: ValidationInfo) -> list[str]:
        globs = []
        for glob in value:
            if glob == "@ith":
                glob = f"{info.data['num']:02}*.in"
            if not glob.endswith(".in"):
                raise PydanticCustomError(
                    "in_globs_end_in", "In_globs must end with '.in'"
                )
            globs.append(glob)

        return globs

    @field_validator("direct_predecessors", mode="before")
    @classmethod
    def expand_predecessors(cls, value: str, info: ValidationInfo) -> list[str]:
        if info.context is None:
            raise RuntimeError(MISSING_VALIDATION_CONTEXT)
        subtask_cnt = info.context.get("subtask_count")
        number = info.data["num"]

        predecessors = []
        for pred in value.split():
            if pred == "@previous":
                if number <= 1:
                    continue
                predecessors.append(number - 1)
            else:
                try:
                    num = int(pred)
                except ValueError:
                    raise PydanticCustomError(
                        "predecessors_must_be_int", "Predecessors must be int"
                    )
                if not 0 <= num < subtask_cnt:
                    raise PydanticCustomError(
                        "predecessors_must_be_in_range",
                        f"Predecessors must be in range 0, {subtask_cnt-1}",
                    )
                predecessors.append(num)

        return list(sorted(set(predecessors)))

    @model_validator(mode="after")
    def validate_model(self):
        if self.name == "@auto":
            self.name = f"Subtask {self.num}"

        return self


class SolutionConfig(BaseEnv):
    """Configuration of one solution."""

    _section: str
    name: str
    primary: bool
    source: str
    points: MaybeInt
    points_above: MaybeInt
    points_below: MaybeInt
    subtasks: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def load_dict(cls, name: ConfigValue, configs: ConfigHierarchy) -> ConfigValuesDict:
        KEYS = [
            "primary",
            "source",
            "points",
            "points_above",
            "points_below",
            "subtasks",
        ]
        args = {
            key: configs.get_from_candidates(
                [(name.section, key), ("all_solutions", key)]
            )
            for key in KEYS
        }
        return {"_section": configs.get(name.section, None), "name": name, **args}

    @field_validator("name", mode="after")
    @classmethod
    def convert_checker(cls, value: str) -> str:
        for banned_char in ".[]":
            if banned_char in value:
                raise PydanticCustomError(
                    "invalid_solution_name",
                    f"Solution name must not contain '{banned_char}'",
                )
        return value

    @field_validator("primary", mode="before")
    @classmethod
    def convert_yesno(cls, value: str, info: ValidationInfo) -> bool:
        if value == "yes":
            return True
        elif value == "no":
            return False
        raise PydanticCustomError(
            "primary_invalid",
            "Must be one of (yes, no)",
        )

    @field_validator("source", mode="before")
    @classmethod
    def convert_auto(cls, value: str, info: ValidationInfo) -> str:
        if value == "@auto":
            return info.data.get("name", "")
        return value

    @field_validator("subtasks", mode="after")
    def validate_subtasks(cls, value, info: ValidationInfo):
        if info.context is None:
            raise RuntimeError(MISSING_VALIDATION_CONTEXT)
        subtask_cnt = info.context.get("subtask_count")
        primary = info.data.get("primary")
        if value == "@auto":
            value = ("1" if primary else "X") * subtask_cnt
        elif value == "@all":
            value = "1" * subtask_cnt
        elif value == "@any":
            value = "X" * subtask_cnt

        if len(value) != subtask_cnt:
            raise PydanticCustomError(
                "subtasks_str_invalid_len",
                f"There are {subtask_cnt} subtasks but subtask string has {len(value)} characters",
            )

        for char in value:
            if char not in SUBTASK_SPEC:
                raise PydanticCustomError(
                    "subtasks_str_invalid_char",
                    f"Not allowed char in subtask string: {char}. Recognized are {''.join(SUBTASK_SPEC.keys())}",
                )

        if primary and value != "1" * subtask_cnt:
            raise PydanticCustomError(
                "primary_sol_must_succeed",
                f"Primary solution must have: subtasks={'1'*subtask_cnt}",
            )

        return value

    @model_validator(mode="after")
    def validate_model(self):
        for points_limit in ["points_above", "points_below"]:
            if self.points is not None and getattr(self, points_limit) is not None:
                raise PydanticCustomError(
                    "points_double_set",
                    f"Both 'points' and '{points_limit}' are set at once",
                    {"points": self.points, points_limit: getattr(self, points_limit)},
                )

        return self


class ProgramLimits(BaseEnv):
    """Configuration of limits of one program type."""

    _section: str = "limits"

    time_limit: float = Field(ge=0)  # [seconds]
    clock_mul: float = Field(ge=0)  # [1]
    clock_min: float = Field(ge=0)  # [seconds]
    mem_limit: int = Field(ge=0)  # [KB]
    process_limit: int = Field(ge=0)
    # limit=0 means unlimited

    def clock_limit(self, override_time_limit: Optional[float] = None) -> float:
        tl = override_time_limit if override_time_limit is not None else self.time_limit
        if tl == 0:
            return 0
        return max(tl * self.clock_mul, self.clock_min)

    @classmethod
    def load_dict(cls, part: ProgramType, configs: ConfigHierarchy) -> ConfigValuesDict:
        def get_limit(limit: str) -> ConfigValue:
            if part == ProgramType.sec_solve:
                return configs.get_from_candidates(
                    [("limits", f"{part.name}_{limit}"), ("limits", f"solve_{limit}")]
                )
            else:
                return configs.get("limits", f"{part.name}_{limit}")

        args: dict[str, Any] = {limit: get_limit(limit) for limit in cls.model_fields}

        return args


class LimitsConfig(BaseEnv):
    """Configuration of limits for all program types."""

    _section: str = "limits"

    tool: ProgramLimits
    in_gen: ProgramLimits
    checker: ProgramLimits
    solve: ProgramLimits
    sec_solve: ProgramLimits
    judge: ProgramLimits
    input_max_size: int
    output_max_size: int

    @classmethod
    def load_dict(cls, configs: ConfigHierarchy) -> ConfigValuesDict:
        args: dict[str, Any] = {}
        for part in ProgramType:
            args[part.name] = ProgramLimits.load_dict(part, configs)

        for file_type in ("input", "output"):
            key = f"{file_type}_max_size"
            args[key] = configs.get("limits", key)

        return args


class CMSConfig(BaseEnv):
    _section: str = "cms"

    title: str
    submission_format: ListStr

    time_limit: float = Field(gt=0)  # [seconds]
    mem_limit: int = Field(gt=0)  # [KB]

    max_submissions: MaybeInt = Field(gt=0)
    min_submission_interval: int = Field(ge=0)  # [seconds]

    score_precision: int = Field(ge=0)
    score_mode: CMSScoreMode
    feedback_level: CMSFeedbackLevel

    @classmethod
    def load_dict(cls, configs: ConfigHierarchy) -> ConfigValuesDict:
        KEYS = [
            "title",
            "submission_format",
            "time_limit",
            "mem_limit",
            "max_submissions",
            "min_submission_interval",
            "score_precision",
            "score_mode",
            "feedback_level",
        ]

        return {key: configs.get("cms", key) for key in KEYS}

    @field_validator("title", mode="before")
    @classmethod
    def convert_title(cls, value: str, info: ValidationInfo) -> str:
        if value == "@name":
            if info.context is None:
                raise RuntimeError(MISSING_VALIDATION_CONTEXT)

            return info.context.get("name")
        else:
            return value

    @field_validator("submission_format", mode="after")
    @classmethod
    def convert_format(cls, value: list[str], info: ValidationInfo) -> list[str]:
        if info.context is None:
            raise RuntimeError(MISSING_VALIDATION_CONTEXT)

        return [
            (
                CMSConfig.get_default_file_name(info.context.get("name"))
                if n == "@name"
                else n
            )
            for n in value
        ]

    @classmethod
    def get_default_file_name(cls, name: str):
        name = re.sub(r"[^a-zA-Z0-9]+", "_", name)
        return f"{name}.%l"


class ChecksConfig(BaseEnv):
    """Configuration of checks for pisek to run."""

    _section: str = "checks"

    checker_distinguishes_subtasks: bool
    solution_for_each_subtask: bool

    @classmethod
    def load_dict(cls, configs: ConfigHierarchy) -> ConfigValuesDict:
        return {key: configs.get("checks", key) for key in cls.model_fields}


def _format_message(err: ErrorDetails) -> str:
    inp = err["input"]
    ctx = err["ctx"] if "ctx" in err else None
    if isinstance(inp, dict) and ctx is not None:
        if ctx == {}:
            return f"{err['msg']}."
        return f"{err['msg']}:\n" + tab(
            "\n".join(f"{key}={val}" for key, val in ctx.items())
        )
    return f"{err['msg']}: '{inp}'"


def _convert_errors(e: ValidationError, config_values: ConfigValuesDict) -> list[str]:
    error_msgs: list[str] = []
    for error in e.errors():
        value: Any = config_values
        for loc in error["loc"]:
            value = value[loc]

        if not isinstance(value, ConfigValue):
            location = (
                value["_section"].location()
                if "_section" in value
                else "Global config error:"
            )
        else:
            location = value.location()

        error_msgs.append(f"In {location}:\n" + tab(_format_message(error)))
    return error_msgs


def load_config(
    path: str,
    strict: bool = False,
    no_colors: bool = False,
    suppress_warnings: bool = False,
    pisek_directory: Optional[str] = None,
) -> Optional[TaskConfig]:
    """Loads config from given path."""
    try:
        config_hierarchy = ConfigHierarchy(path, no_colors, pisek_directory)
        config_values = TaskConfig.load_dict(config_hierarchy)
        config = TaskConfig(**_to_values(config_values))
        config_hierarchy.check_unused_keys()
        if config_hierarchy.check_todos() and not suppress_warnings:
            warn("Unsolved TODOs in config.", TaskConfigError, strict, no_colors)
        return config
    except TaskConfigError as err:
        eprint(colored(str(err), "red", no_colors))
    except ValidationError as err:
        eprint(
            colored(
                "Invalid config:\n\n"
                + "\n\n".join(_convert_errors(err, config_values)),
                "red",
                no_colors,
            )
        )
    return None



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import deque

from pisek.jobs.job_pipeline import JobPipeline
from pisek.env.env import Env, TestingTarget

from pisek.jobs.parts.task_job import (
    TOOLS_MAN_CODE,
    INPUTS_MAN_CODE,
    GENERATOR_MAN_CODE,
    CHECKER_MAN_CODE,
    JUDGE_MAN_CODE,
    SOLUTION_MAN_CODE,
    DATA_MAN_CODE,
)

from pisek.jobs.parts.tools import ToolsManager
from pisek.jobs.parts.data import DataManager
from pisek.jobs.parts.generator import GeneratorManager
from pisek.jobs.parts.checker import CheckerManager
from pisek.jobs.parts.judge import JudgeManager
from pisek.jobs.parts.solution import SolutionManager
from pisek.jobs.parts.data import DataCheckingManager
from pisek.jobs.parts.testing_log import CreateTestingLog
from pisek.jobs.parts.completeness_check import CompletenessCheck


class TaskPipeline(JobPipeline):
    """JobPipeline that checks whether task behaves as expected."""

    def __init__(self, env: Env):
        super().__init__()
        named_pipeline = [
            tools := (ToolsManager(), TOOLS_MAN_CODE),
            generator := (GeneratorManager(), GENERATOR_MAN_CODE),
            inputs := (DataManager(), INPUTS_MAN_CODE),
        ]
        generator[0].add_prerequisite(*tools)
        inputs[0].add_prerequisite(*generator)

        if env.target != "solution":
            named_pipeline.append(checker := (CheckerManager(), CHECKER_MAN_CODE))
            checker[0].add_prerequisite(*inputs)

        solutions = []
        if env.solutions:
            named_pipeline.append(judge := (JudgeManager(), JUDGE_MAN_CODE))
            judge[0].add_prerequisite(*inputs)

            if env.config.judge_needs_out or (
                env.config.primary_solution in env.solutions
            ):
                named_pipeline.append(
                    primary_solution := (
                        SolutionManager(env.config.primary_solution),
                        f"{SOLUTION_MAN_CODE}{env.config.primary_solution}",
                    )
                )
                solutions.append(primary_solution)

        for sol_name in env.solutions:
            if sol_name == env.config.primary_solution:
                continue
            named_pipeline.append(
                solution := (
                    SolutionManager(sol_name),
                    f"{SOLUTION_MAN_CODE}{sol_name}",
                )
            )
            if env.config.judge_needs_out:
                solution[0].add_prerequisite(*primary_solution)
            solutions.append(solution)

        for solution in solutions:
            solution[0].add_prerequisite(*inputs)
            solution[0].add_prerequisite(*judge)

        named_pipeline.append(data_check := (DataCheckingManager(), DATA_MAN_CODE))
        data_check[0].add_prerequisite(*inputs)
        for solution in solutions:
            data_check[0].add_prerequisite(*solution)

        if env.testing_log:
            named_pipeline.append(testing_log := (CreateTestingLog(), ""))
            for solution in solutions:
                testing_log[0].add_prerequisite(*solution)

        if env.target == TestingTarget.all:
            named_pipeline.append(completeness_check := (CompletenessCheck(), ""))
            for solution in solutions:
                completeness_check[0].add_prerequisite(*solution)

        self.pipeline = deque(map(lambda x: x[0], named_pipeline))



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Optional

from pisek.jobs.jobs import State, Job, PipelineItemFailure
from pisek.env.env import Env
from pisek.utils.paths import TaskPath
from pisek.config.task_config import ProgramType
from pisek.config.config_types import Scoring
from pisek.utils.terminal import colored_env
from pisek.jobs.parts.task_job import TaskJobManager
from pisek.jobs.parts.run_result import RunResult, RunResultKind
from pisek.jobs.parts.program import ProgramsJob
from pisek.jobs.parts.compile import Compile


class CheckerManager(TaskJobManager):
    """Runs checker on inputs."""

    def __init__(self):
        self.skipped_checker = ""
        super().__init__("Running checker")

    def _get_jobs(self) -> list[Job]:
        if self._env.config.checker is None:
            if self._env.strict:
                raise PipelineItemFailure("No checker specified in config.")
            else:
                self.skipped_checker = colored_env(
                    "Warning: No checker specified in config.\n"
                    "It is recommended to set `checker` is section [tests]",
                    "yellow",
                    self._env,
                )
            return []

        checker = TaskPath(self._env.config.checker)

        jobs: list[Job] = [compile_checker := Compile(self._env, checker)]

        self.loose_subtasks = []
        for sub_num, sub in self._env.config.subtasks.items():
            if sub_num == 0:
                continue  # Skip samples
            for inp in self._subtask_inputs(sub):
                jobs.append(
                    check := CheckerJob(
                        self._env, checker, inp, sub_num, RunResultKind.OK
                    )
                )
                check.add_prerequisite(compile_checker)
            if self._env.config.checks.checker_distinguishes_subtasks:
                if sub.direct_predecessors:
                    self.loose_subtasks.append(LooseCheckJobGroup(sub_num))
                    for pred in sub.direct_predecessors:
                        if pred == 0:
                            continue  # Skip samples
                        self.loose_subtasks[-1].jobs[pred] = []
                        for inp in self._subtask_new_inputs(sub):
                            jobs.append(
                                check := CheckerJob(self._env, checker, inp, pred, None)
                            )
                            self.loose_subtasks[-1].jobs[pred].append(check)
                            check.add_prerequisite(compile_checker)

        return jobs

    def _evaluate(self) -> None:
        if len(self.jobs) == 0:
            return

        for loose_subtask in self.loose_subtasks:
            err = loose_subtask.failed(self._env.config.scoring)
            if err is not None:
                raise PipelineItemFailure(err)

    def _get_status(self) -> str:
        if self.skipped_checker:
            if self.state == State.succeeded:
                return self.skipped_checker
            else:
                return ""
        else:
            return super()._get_status()


class LooseCheckJobGroup:
    """
    Groups jobs on subtask where checker is run on predecessors instead.
    Checking that checker is strict enough - checking fails when run on predecessor.
    """

    def __init__(self, num: int):
        self.num = num
        self.jobs: dict[int, list[CheckerJob]] = {}

    def failed(self, scoring: Scoring) -> Optional[str]:
        """Returns whether jobs resulted as expected."""

        def result_kind(job: CheckerJob) -> RunResultKind:
            if job.result is None:
                raise RuntimeError(f"Job {job.name} has not finished yet.")
            return job.result.kind

        for pred in self.jobs:
            results = list(map(result_kind, self.jobs[pred]))
            if scoring == Scoring.equal and RunResultKind.OK in results:
                job = self._index_job(pred, results, RunResultKind.OK)
                return (
                    f"Checker is not strict enough:\n"
                    f"All inputs of subtask {self.num} should have not passed on predecessor subtask {pred}\n"
                    f"but on input {job.input} did not."
                )
            if scoring == Scoring.min and RunResultKind.RUNTIME_ERROR not in results:
                return (
                    f"Checker is not strict enough:\n"
                    f"An input of subtask {self.num} should have not passed on predecessor subtask {pred}\n"
                    f"but all have passed."
                )
            if RunResultKind.TIMEOUT in results:
                to_job = self._index_job(pred, results, RunResultKind.TIMEOUT)
                return f"Checker timeouted on input {to_job.input}, subtask {self.num}."
        return None

    def _index_job(
        self, pred: int, results: list[RunResultKind], result: RunResultKind
    ) -> "CheckerJob":
        return self.jobs[pred][results.index(result)]


class CheckerJob(ProgramsJob):
    """Runs checker on single input."""

    def __init__(
        self,
        env: Env,
        checker: TaskPath,
        input_: TaskPath,
        subtask: int,
        expected: Optional[RunResultKind],
        **kwargs,
    ):
        super().__init__(
            env=env, name=f"Check {input_:n} on subtask {subtask}", **kwargs
        )
        self.checker = checker
        self.subtask = subtask
        self.input = input_
        self.log_file = TaskPath.log_file(
            self._env, input_.name, f"{self.checker.name}{subtask}"
        )
        self.expected = expected

    def _check(self) -> RunResult:
        return self._run_program(
            ProgramType.checker,
            self.checker,
            args=[str(self.subtask)],
            stdin=self.input,
            stderr=self.log_file,
        )

    def _run(self) -> RunResult:
        result = self._check()
        if self.expected == RunResultKind.OK != result.kind:
            raise self._create_program_failure(
                f"Checker failed on {self.input:p} (subtask {self.subtask}) but should have succeeded.",
                result,
            )
        elif self.expected == RunResultKind.RUNTIME_ERROR != result.kind:
            raise self._create_program_failure(
                f"Checker succeeded on {self.input:p} (subtask {self.subtask}) but should have failed.",
                result,
            )
        return result



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
from typing import Any, Callable

from pisek.config.config_types import DataFormat
from pisek.jobs.jobs import Job, PipelineItemFailure
from pisek.env.env import Env
from pisek.utils.paths import TaskPath
from pisek.jobs.parts.task_job import (
    TaskJob,
    TaskJobManager,
    GENERATOR_MAN_CODE,
    INPUTS_MAN_CODE,
    SOLUTION_MAN_CODE,
)
from pisek.jobs.parts.solution_result import Verdict
from pisek.jobs.parts.tools import IsClean


class DataManager(TaskJobManager):
    """Moves data to correct folders."""

    def __init__(self) -> None:
        self._static_inputs: list[TaskPath] = []
        self._generated_inputs: list[TaskPath] = []
        self._static_outputs: list[TaskPath] = []
        super().__init__("Moving data")

    def _get_jobs(self) -> list[Job]:
        self._static_inputs = self.globs_to_files(
            self._env.config.input_globs, TaskPath.static_path(self._env, ".")
        )
        self._static_outputs = self.globs_to_files(
            map(
                lambda g: re.sub(r"\.in$", ".out", g),
                self._env.config.input_globs,
            ),
            TaskPath.static_path(self._env, "."),
        )
        self._generated_inputs = self.prerequisites_results[GENERATOR_MAN_CODE][
            "inputs"
        ]

        all_input_files = self._static_inputs + self._generated_inputs
        all_output_files = self._static_outputs
        jobs: list[Job] = []

        if self._env.config.task_type != "communication":
            for static_inp in self._static_inputs:
                static_out = static_inp.replace_suffix(".out")
                if static_out not in self._static_outputs:
                    raise PipelineItemFailure(
                        f"Missing matching output '{static_out:p}' for static input '{static_inp:p}'."
                    )

        self._inputs: list[TaskPath] = []
        self._outputs: list[TaskPath] = []
        link: LinkData
        for fname in all_input_files:
            jobs.append(link := LinkInput(self._env, fname))
            self._inputs.append(link.dest)
        for fname in all_output_files:
            jobs.append(link := LinkOutput(self._env, fname))
            self._outputs.append(link.dest)

        for num, sub in self._env.config.subtasks.items():
            inputs = list(filter(lambda p: sub.in_subtask(p.name), all_input_files))
            if len(inputs) == 0:
                raise PipelineItemFailure(
                    f"No inputs for subtask {num} with globs {sub.all_globs}."
                )

        return jobs

    def _compute_result(self) -> dict[str, Any]:
        res = {
            "inputs": self._inputs,
            "outputs": self._outputs,
            "all": self._inputs + self._outputs,
        }

        return res


class DataJob(TaskJob):
    def __init__(self, env: Env, name: str, data: TaskPath, **kwargs) -> None:
        super().__init__(
            env=env,
            name=name,
            **kwargs,
        )
        self.data = data


class LinkData(DataJob):
    """Copy data to into dest folder."""

    def __init__(self, env: Env, data: TaskPath, dest: TaskPath, **kwargs) -> None:
        super().__init__(
            env=env, name=f"Copy {data:p} to {dest:p}/", data=data, **kwargs
        )
        self.dest = TaskPath(dest.path, self.data.name)

    def _run(self):
        self._link_file(self.data, self.dest, overwrite=True)


class LinkInput(LinkData):
    """Copy input to its place."""

    def __init__(self, env: Env, input_: TaskPath, **kwargs) -> None:
        super().__init__(
            env=env, data=input_, dest=TaskPath.input_path(self._env, "."), **kwargs
        )


class LinkOutput(LinkData):
    """Copy output to its place."""

    def __init__(self, env: Env, output: TaskPath, **kwargs) -> None:
        super().__init__(
            env=env, data=output, dest=TaskPath.output_path(self._env, "."), **kwargs
        )


MB = 1024 * 1024


class DataCheckingManager(TaskJobManager):
    def __init__(self):
        super().__init__("Checking data")

    def _get_jobs(self) -> list[Job]:
        jobs: list[Job] = []

        inputs = []
        outputs = []
        for name, data in self.prerequisites_results.items():
            if name.startswith(INPUTS_MAN_CODE):
                inputs += data["inputs"]
            elif name.startswith(SOLUTION_MAN_CODE):
                outs = data["outputs"]
                outputs += (
                    outs[Verdict.ok]
                    + outs[Verdict.partial_ok]
                    + outs[Verdict.wrong_answer]
                )
        for inp in inputs:
            if self._env.config.in_format == DataFormat.text:
                jobs.append(IsClean(self._env, inp))

            if self._env.config.limits.input_max_size != 0:
                jobs.append(InputSmall(self._env, inp))

        for out in outputs:
            if self._env.config.out_format == DataFormat.text:
                jobs.append(IsClean(self._env, out))

            if self._env.config.limits.output_max_size != 0:
                jobs.append(OutputSmall(self._env, out))

        return jobs


class InputSmall(DataJob):
    """Checks that input is small enough to download."""

    def __init__(self, env: Env, input_: TaskPath, **kwargs) -> None:
        super().__init__(
            env=env,
            name=f"Input {input_:n} is smaller than {env.config.limits.input_max_size}MB",
            data=input_,
            **kwargs,
        )

    def _run(self):
        max_size = self._env.config.limits.input_max_size
        if (sz := self._file_size(self.data)) > max_size * MB:
            raise PipelineItemFailure(
                f"Input {self.data:p} is bigger than {max_size}MB: {(sz+MB-1)//MB}MB"
            )


class OutputSmall(DataJob):
    """Checks that output is small enough to upload."""

    def __init__(self, env: Env, output: TaskPath, **kwargs) -> None:
        super().__init__(
            env=env,
            name=f"Output {output:n} is smaller than {env.config.limits.output_max_size}MB",
            data=output,
            **kwargs,
        )

    def _run(self):
        max_size = self._env.config.limits.output_max_size
        if (sz := self._file_size(self.data)) > max_size * MB:
            raise PipelineItemFailure(
                f"Output {self.data} is bigger than {max_size}MB: {(sz+MB-1)//MB}MB"
            )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import re
import shutil
from typing import Optional

from pisek.jobs.cache import CACHE_FILENAME
from pisek.config.task_config import TaskConfig, load_config

BUILD_DIR = "build/"


def rm_f(fn):
    try:
        os.unlink(fn)
    except FileNotFoundError:
        pass


def get_build_dir(task_dir: str) -> str:
    return os.path.normpath(os.path.join(task_dir, BUILD_DIR))


def pad_num(what: int, length=2):
    return f"{'0'*(length-len(str(what)))}{what}"


def get_input_name(seed: int, subtask: int) -> str:
    return f"{pad_num(subtask)}_{seed:x}.in"


def get_output_name(input_file: str, solution_name: str) -> str:
    """
    >>> get_output_name("sample.in", "solve_6b")
    'sample.solve_6b.out'
    """
    no_suffix = os.path.splitext(os.path.basename(input_file))[0]
    if solution_name == "":
        return f"{no_suffix}.out"
    return f"{no_suffix}.{os.path.basename(solution_name)}.out"


def _clean_subdirs(task_dir: str, subdirs: list[str]) -> None:
    for subdir in subdirs:
        full = os.path.join(task_dir, subdir)
        try:
            shutil.rmtree(full)
        except FileNotFoundError:
            pass


def clean_task_dir(task_dir: str, pisek_directory: Optional[str]) -> bool:
    config = load_config(
        task_dir, suppress_warnings=True, pisek_directory=pisek_directory
    )
    if config is None:
        return False
    # XXX: ^ safeguard, raises an exception if task_dir isn't really a task
    # directory

    rm_f(os.path.join(task_dir, CACHE_FILENAME))
    _clean_subdirs(task_dir, [config.data_subdir, BUILD_DIR])
    return True


def get_expected_score(solution_name: str, config: TaskConfig) -> Optional[int]:
    """
    solve -> 10 (assuming 10 is the maximum score)
    solve_0b -> 0
    solve_jirka_4b -> 4
    """
    match = re.search(r"_([0-9]{1,3}|X)b$", solution_name)

    if match:
        if match[1] == "X":
            return None
        score = int(match[1])
        return score
    else:
        return config.total_points


def quote_output(s, max_length=1500, max_lines=20):
    """
    Indicates that a string is a quote of another program's output by adding
    indentation and color.
    """
    if isinstance(s, bytes):
        s = s.decode("utf-8")

    while s and s[-1] == "\n":
        s = s[:-1]

    add_ellipsis = False

    if len(s) > max_length:
        s = s[:max_length]
        add_ellipsis = True

    lines = s.split("\n")
    if len(lines) > max_lines:
        lines = lines[:max_lines]
        add_ellipsis = True

    s = "\n".join(("  " + l) for l in lines)

    if add_ellipsis:
        s += "\n  [...]"

    return s



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import glob
import random
import os
import shutil
from typing import Any

from pisek.env.env import Env
from pisek.utils.paths import TaskPath, GENERATED_SUBDIR
from pisek.config.task_config import ProgramType
from pisek.jobs.jobs import Job, PipelineItemFailure
from pisek.jobs.parts.task_job import TaskJob, TaskJobManager
from pisek.jobs.parts.run_result import RunResult, RunResultKind
from pisek.jobs.parts.program import ProgramsJob
from pisek.jobs.parts.compile import Compile


class GeneratorManager(TaskJobManager):
    """Manager that generates inputs and test generator."""

    def __init__(self):
        self._inputs = []
        super().__init__("Running generator")

    def _get_jobs(self) -> list[Job]:
        generator = TaskPath(self._env.config.in_gen)

        jobs: list[Job] = [compile_gen := Compile(self._env, generator)]

        if self._env.config.contest_type == "kasiopea":
            random.seed(4)  # Reproducibility!
            seeds = random.sample(range(0, 16**4), self._env.inputs)
            for sub_num, _ in self._env.config.subtasks.items():
                if sub_num == 0:
                    continue  # skip samples
                last_gen: OnlineGeneratorGenerate
                for i, seed in enumerate(seeds):
                    self._inputs.append(
                        input_ := TaskPath.generated_input_file(
                            self._env, int(sub_num), seed
                        )
                    )

                    jobs.append(
                        gen := OnlineGeneratorGenerate(
                            self._env, generator, input_, sub_num, seed
                        )
                    )
                    gen.add_prerequisite(compile_gen)
                    if i == 0:
                        jobs.append(
                            det := OnlineGeneratorDeterministic(
                                self._env, generator, input_, sub_num, seed
                            )
                        )
                        det.add_prerequisite(gen)
                    elif i == 1:
                        jobs.append(
                            rs := OnlineGeneratorRespectsSeed(
                                self._env,
                                sub_num,
                                last_gen.seed,
                                gen.seed,
                                last_gen.input_,
                                gen.input_,
                            )
                        )
                        rs.add_prerequisite(last_gen)
                        rs.add_prerequisite(gen)
                    last_gen = gen
        else:
            jobs.append(gen2 := OfflineGeneratorGenerate(self._env, generator))
            gen2.add_prerequisite(compile_gen)

        return jobs

    def _compute_result(self) -> dict[str, Any]:
        res = {}
        if self._env.config.contest_type == "kasiopea":
            res["inputs"] = self._inputs
        else:
            res["inputs"] = self.globs_to_files(
                self._env.config.input_globs,
                TaskPath.generated_path(self._env, "."),
            )

        return res


class RunOnlineGeneratorMan(TaskJobManager):
    def __init__(self, subtask: int, seed: int, file: str):
        self._subtask = subtask
        self._seed = seed
        self._file = file
        super().__init__("Running generator")

    def _get_jobs(self) -> list[Job]:
        generator = TaskPath(self._env.config.in_gen)

        jobs: list[Job] = [
            compile_gen := Compile(self._env, generator),
            gen := OnlineGeneratorGenerate(
                self._env,
                generator,
                TaskPath(self._file),
                self._subtask,
                self._seed,
            ),
        ]
        gen.add_prerequisite(compile_gen)

        return jobs


class OnlineGeneratorJob(ProgramsJob):
    """Abstract class for jobs with OnlineGenerator."""

    def __init__(
        self,
        env: Env,
        name: str,
        generator: TaskPath,
        input_: TaskPath,
        subtask: int,
        seed: int,
        **kwargs,
    ) -> None:
        super().__init__(env=env, name=name, **kwargs)
        self.generator = generator
        self.subtask = subtask
        self.seed = seed
        self.input_ = input_

    def _gen(self, input_file: TaskPath, seed: int, subtask: int) -> RunResult:
        if seed < 0:
            raise PipelineItemFailure(f"Seed {seed} is negative.")

        input_dir = os.path.dirname(input_file.path)

        difficulty = str(subtask)
        hexa_seed = f"{seed:x}"

        result = self._run_program(
            ProgramType.in_gen,
            self.generator,
            args=[difficulty, hexa_seed],
            stdout=input_file,
            stderr=TaskPath.log_file(self._env, self.input_.name, self.generator.name),
        )
        if result.kind != RunResultKind.OK:
            raise self._create_program_failure(
                f"{self.generator} failed on subtask {subtask}, seed {seed:x}:", result
            )

        return result


class OnlineGeneratorGenerate(OnlineGeneratorJob):
    """Generates single input using OnlineGenerator."""

    def __init__(
        self,
        env: Env,
        generator: TaskPath,
        input_: TaskPath,
        subtask: int,
        seed: int,
        **kwargs,
    ) -> None:
        super().__init__(
            env=env,
            name=f"Generate {input_.name}",
            generator=generator,
            input_=input_,
            subtask=subtask,
            seed=seed,
            **kwargs,
        )

    def _run(self) -> RunResult:
        return self._gen(self.input_, self.seed, self.subtask)


class OnlineGeneratorDeterministic(OnlineGeneratorJob):
    """Test whether generating given input again has same result."""

    def __init__(
        self,
        env: Env,
        generator: TaskPath,
        input_: TaskPath,
        subtask: int,
        seed: int,
        **kwargs,
    ) -> None:
        super().__init__(
            env=env,
            name=f"Generator is deterministic (subtask {subtask}, seed {seed:x})",
            generator=generator,
            input_=input_,
            subtask=subtask,
            seed=seed,
            **kwargs,
        )

    def _run(self) -> None:
        copy_file = self.input_.replace_suffix(".in2")
        self._gen(copy_file, self.seed, self.subtask)
        if not self._files_equal(self.input_, copy_file):
            raise PipelineItemFailure(
                f"Generator is not deterministic. Files {self.input_:p} and {copy_file:p} differ "
                f"(subtask {self.subtask}, seed {self.seed})",
            )


class OnlineGeneratorRespectsSeed(TaskJob):
    """Test whether two files generated with different seed are different."""

    def __init__(
        self,
        env: Env,
        subtask: int,
        seed1: int,
        seed2: int,
        file1: TaskPath,
        file2: TaskPath,
        **kwargs,
    ) -> None:
        self.file1, self.file2 = file1, file2
        self.subtask = subtask
        self.seed1, self.seed2 = seed1, seed2
        super().__init__(
            env=env,
            name=f"Generator respects seeds ({self.file1:n} and {self.file2:n} are different)",
            **kwargs,
        )

    def _run(self) -> None:
        if self._files_equal(self.file1, self.file2):
            raise PipelineItemFailure(
                f"Generator doesn't respect seed."
                f"Files {self.file1:n} (seed {self.seed1:x}) and {self.file2:n} (seed {self.seed2:x}) are same."
            )


class OfflineGeneratorGenerate(ProgramsJob):
    """Job that generates all inputs using OfflineGenerator."""

    def __init__(self, env: Env, generator: TaskPath, **kwargs) -> None:
        super().__init__(env=env, name="Generate inputs", **kwargs)
        self.generator = generator

    def _gen(self) -> None:
        """Generates all inputs."""
        gen_dir = TaskPath.generated_path(self._env, ".")
        try:
            shutil.rmtree(gen_dir.path)
        except FileNotFoundError:
            pass
        self.makedirs(gen_dir, exist_ok=False)

        run_result = self._run_program(
            ProgramType.in_gen,
            self.generator,
            args=[gen_dir.path],
            stderr=TaskPath.log_file(self._env, None, self.generator.name),
        )
        self._access_file(gen_dir)

        if run_result.kind != RunResultKind.OK:
            raise self._create_program_failure("Generator failed:", run_result)

    def _run(self) -> None:
        self._gen()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from enum import StrEnum, auto
from pydantic import Field
from typing import Any, TYPE_CHECKING, Callable, TypeVar


from pisek.env.context import ContextModel


T = TypeVar("T")
TFunc = Callable[..., T]


class BaseEnv(ContextModel):
    """
    Collection of environment variables which logs whether each variable was accessed.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._accessed: set[str] = set()
        self._logging: bool = True
        self._locked: bool = False

    if not TYPE_CHECKING:

        def __getattribute__(self, item: str) -> Any:
            """Overloaded __getattribute__ that logs accesses to fields"""
            # Implementing this method is kind of magical and dangerous. Beware!
            if (
                not item.startswith("_")
                and (item != "model_fields" and item != "model_computed_fields")
                and "_accessed" in self.__dict__
                and self._logging
                and (item in self.model_fields or item in self.model_computed_fields)
            ):
                self._accessed.add(item)
            return super().__getattribute__(item)

    def fork(self):
        """Make copy of this env with no accesses logged."""
        if self._locked:
            raise RuntimeError("Locked BaseEnv cannot be forked.")

        model = self.model_copy(deep=True)
        model._clear_accesses()
        return model

    @staticmethod
    def _recursive_call(
        function: Callable[["BaseEnv"], None]
    ) -> Callable[["BaseEnv"], None]:
        def recursive(self: "BaseEnv") -> None:
            self._logging = False
            for key in self.model_fields:
                item = getattr(self, key)
                if isinstance(item, BaseEnv):
                    recursive(item)
                elif isinstance(item, dict):
                    for subitem in item.values():
                        if isinstance(subitem, BaseEnv):
                            recursive(subitem)

            self._logging = True
            function(self)

        return recursive

    @_recursive_call
    def _clear_accesses(self) -> None:
        """Remove all logged accesses."""
        self._accessed = set()

    @_recursive_call
    def lock(self) -> None:
        """Prevent this Env (and subenvs) from being forked."""
        self._locked = True

    def get_accessed(self) -> list[str]:
        """Get all accessed field names in this env (and all subenvs)."""
        accessed = []
        self._logging = False
        for key in self._accessed:
            item = getattr(self, key)
            if isinstance(item, BaseEnv):
                accessed += [(f"{key}.{subkey}") for subkey in item.get_accessed()]
            elif isinstance(item, dict) and all(
                isinstance(val, BaseEnv) for val in item.values()
            ):
                accessed += [
                    (f"{key}[{dict_key}].{subkey}")
                    for dict_key, subenv in item.items()
                    for subkey in subenv.get_accessed()
                ]
            else:
                accessed.append(key)
        self._logging = True

        return accessed

    def get_compound(self, key: str) -> Any:
        """Get attribute that may be nested deeper and indexed."""
        obj = self
        for key_part in key.split("."):
            if "[" in key_part:
                assert key_part[-1] == "]"
                var, index = key_part[:-1].split("[")
                dict_ = getattr(obj, var)
                obj = dict_[type(list(dict_.keys())[0])(index)]
            else:
                obj = getattr(obj, key_part)
        return obj



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from importlib.resources import files
from typing import Optional

import subprocess
from pisek.jobs.jobs import State, Job, PipelineItemFailure
from pisek.env.env import Env
from pisek.utils.paths import TaskPath
from pisek.config.task_config import ProgramType
from pisek.jobs.parts.task_job import TaskJob, TaskJobManager
from pisek.jobs.parts.program import ProgramsJob


class ToolsManager(TaskJobManager):
    """Manager that prepares all tools necessary for task testing."""

    def __init__(self):
        super().__init__("Preparing tools")

    def _get_jobs(self) -> list[Job]:
        self.makedirs(TaskPath.executable_path(self._env, "."))
        jobs: list[Job] = [
            PrepareMinibox(self._env),
            PrepareTextPreprocessor(self._env),
        ]
        return jobs


class PrepareMinibox(TaskJob):
    """Compiles minibox."""

    def __init__(self, env: Env, **kwargs) -> None:
        super().__init__(env=env, name="Prepare Minibox", **kwargs)

    def _run(self):
        source = files("pisek").joinpath("tools/minibox.c")
        executable = TaskPath.executable_path(self._env, "minibox")
        self._access_file(executable)
        gcc = subprocess.run(
            [
                "gcc",
                source,
                "-o",
                executable.path,
                "-std=gnu11",
                "-D_GNU_SOURCE",
                "-O2",
                "-Wall",
                "-Wextra",
                "-Wno-parentheses",
                "-Wno-sign-compare",
                "-Wno-unused-result",
            ]
        )
        if gcc.returncode != 0:
            raise PipelineItemFailure("Minibox compilation failed.")


class PrepareTextPreprocessor(TaskJob):
    """Compiles Text Preprocessor."""

    def __init__(self, env: Env, **kwargs) -> None:
        super().__init__(env=env, name="Prepare text preprocessor", **kwargs)

    def _run(self):
        source = files("pisek").joinpath("tools/text-preproc.c")
        executable = TaskPath.executable_path(self._env, "text-preproc")
        self._access_file(executable)
        gcc = subprocess.run(
            [
                "gcc",
                source,
                "-o",
                executable.path,
                "-std=gnu11",
                "-O2",
                "-Wall",
                "-Wextra",
                "-Wno-parentheses",
                "-Wno-sign-compare",
            ]
        )
        if gcc.returncode != 0:
            raise PipelineItemFailure("Text preprocessor compilation failed.")


class SanitizeAbstract(ProgramsJob):
    """Abstract job that has method for file sanitization."""

    def _sanitize(self, input_: TaskPath, output: TaskPath) -> None:
        result = self._run_program(
            ProgramType.tool,
            TaskPath.executable_path(self._env, "text-preproc"),
            stdin=input_,
            stdout=output,
        )
        if result.returncode == 43:
            raise self._create_program_failure(
                f"Text preprocessor failed on file: {input_:p}", result
            )


class Sanitize(SanitizeAbstract):
    """Sanitize text file using Text Preprocessor."""

    def __init__(
        self, env: Env, input_: TaskPath, output: Optional[TaskPath] = None, **kwargs
    ) -> None:
        self.input = input_
        self.output = (
            output if output else TaskPath.sanitized_file(self._env, input_.path)
        )
        super().__init__(
            env=env, name=f"Sanitize {self.input:n} -> {self.output:n}", **kwargs
        )

    def _run(self):
        return self._sanitize(self.input, self.output)


class IsClean(SanitizeAbstract):
    """Check that file is same after sanitizing with Text Preprocessor."""

    def __init__(
        self, env: Env, input_: TaskPath, output: Optional[TaskPath] = None, **kwargs
    ) -> None:
        self.input = input_
        self.output = (
            output if output else TaskPath.sanitized_file(self._env, input_.path)
        )
        super().__init__(env=env, name=f"{self.input:n} is clean", **kwargs)

    def _run(self):
        self._sanitize(self.input, self.output)
        if not self._files_equal(self.input, self.output):
            raise PipelineItemFailure(
                f"File {self.input:n} is not clean. Check encoding, missing newline at the end or \\r."
            )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
import subprocess
import sys

from pisek.env.env import Env
from pisek.utils.paths import TaskPath
from pisek.jobs.jobs import State, PipelineItemFailure
from pisek.jobs.parts.program import ProgramsJob


class Compile(ProgramsJob):
    """Job that compiles a program."""

    def __init__(
        self,
        env: Env,
        program: TaskPath,
        use_stub: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(env=env, name=f"Compile {program:n}", **kwargs)
        self.program = program
        self.use_stub = use_stub
        self.target = TaskPath.executable_file(self._env, program.name)

        self.stub = None
        self.headers = []

        if use_stub and self._env.config.stub:
            self.stub = TaskPath(self._env.config.stub)
        if use_stub and self._env.config.headers:
            self.headers = [TaskPath(header) for header in self._env.config.headers]

    def _resolve_extension(self, name: TaskPath) -> TaskPath:
        """
        Given `name`, finds a file named `name`.[ext],
        where [ext] is a file extension for one of the supported languages.

        If a name with a valid extension is given, it is returned unchanged
        """
        extensions = supported_extensions()
        candidates = []
        path = name.path
        for ext in extensions:
            if os.path.isfile(path + ext):
                candidates.append(path + ext)
            if path.endswith(ext) and os.path.isfile(path):
                # Extension already present in `name`
                candidates.append(path)

        if len(candidates) == 0:
            raise PipelineItemFailure(f"No program with given name exists: {path}")
        if len(candidates) > 1:
            raise PipelineItemFailure(
                f"Multiple programs with same name exist: {', '.join(candidates)}"
            )

        return TaskPath.from_abspath(candidates[0])

    def _run(self):
        """Compiles program."""
        program = self._resolve_extension(self.program)
        self.makedirs(TaskPath.executable_path(self._env, "."))

        _, ext = os.path.splitext(program.path)

        if ext in COMPILE_RULES:
            COMPILE_RULES[ext](self, program)
        else:
            raise PipelineItemFailure(f"No rule for compiling {program:p}.")

        self._access_file(program)
        self._access_file(self._load_compiled(self.program))

    def _compile_cpp(self, program: TaskPath):
        cpp_flags = ["-std=c++17", "-O2", "-Wall", "-lm", "-Wshadow", self._c_colors()]

        cpp_flags += self._add_stub("cpp")

        return self._run_compilation(
            ["g++", program.path, "-o", self.target.path] + cpp_flags, program
        )

    def _compile_c(self, program: TaskPath):
        c_flags = ["-std=c17", "-O2", "-Wall", "-lm", "-Wshadow", self._c_colors()]

        c_flags += self._add_stub("c")

        return self._run_compilation(
            ["gcc", program.path, "-o", self.target.path] + c_flags, program
        )

    def _c_colors(self):
        if not self._env.no_colors:
            return "-fdiagnostics-color=always"
        else:
            return "-fdiagnostics-color=never"

    def _compile_pas(self, program: TaskPath):
        build_dir = TaskPath.executable_path(self._env, ".")
        pas_flags = [
            "-gl",
            "-O3",
            "-Sg",
            "-o" + self.target.path,
            "-FE" + build_dir.path,
        ]

        return self._run_compilation(["fpc"] + pas_flags + [program.path], program)

    def _compile_rust(self, program: TaskPath):
        return self._run_compilation(
            ["rustc", "-O", "-o", self.target.path, program.path], program
        )

    def _run_compilation(self, args: list[str], program: TaskPath, **kwargs) -> None:
        self._check_tool(args[0])

        comp = subprocess.Popen(args, **kwargs, stderr=subprocess.PIPE)
        while comp.stderr is not None:
            line = comp.stderr.readline().decode()
            if not line:
                break
            self._print(line, end="", stderr=True)

        comp.wait()
        if comp.returncode != 0:
            raise PipelineItemFailure(f"Compilation of {program:p} failed.")

    def _check_tool(self, tool: str) -> None:
        """Checks that a tool exists."""
        try:
            subprocess.run(
                tool.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=0
            )
        except subprocess.TimeoutExpired:
            pass
        except FileNotFoundError:
            raise PipelineItemFailure(f"Missing tool: {tool}")

    def _compile_script(self, program: TaskPath) -> None:
        if not self.valid_shebang(program):
            raise PipelineItemFailure(
                f"{program} has invalid shebang. "
                "For Python should first line be '#!/usr/bin/env python3' or similar.\n"
                "Check also that you are using linux eol."
            )

        with open(program.path, "r", newline="\n") as f:
            interpreter = f.readline().strip().lstrip("#!")
        self._check_tool(interpreter)

        shutil.copyfile(program.path, self.target.path)
        self._chmod_exec(self.target)

    @staticmethod
    def valid_shebang(program: TaskPath) -> bool:
        """Check if file has shebang and if the shebang is valid"""

        with open(program.path, "r", newline="\n") as f:
            first_line = f.readline()

        if not first_line.startswith("#!"):
            return False

        if first_line.endswith("\r\n"):
            return False

        if os.path.splitext(program.path)[1] == ".py":
            return any(
                first_line == f"#!/usr/bin/env {interpreter}\n"
                for interpreter in VALID_PYTHON_INTERPRETERS
            )
        else:
            # No check performed for non-Python at the moment
            return True

    @staticmethod
    def _chmod_exec(filepath: TaskPath) -> None:
        st = os.stat(filepath.path)
        os.chmod(filepath.path, st.st_mode | 0o111)

    def _add_stub(self, extension):
        flags = []

        if self.stub is not None:
            filename = TaskPath(f"{self.stub:p}.{extension}")
            self._access_file(filename)
            flags += [filename.path]

        for header in self.headers:
            self._access_file(header)
            directory = os.path.normpath(os.path.dirname(header.path))
            flags += [f"-iquote{directory}"]

        return flags


# This is the list of supported Python interpreters.
# Used for checking the shebang.
VALID_PYTHON_INTERPRETERS = ["python3", "pypy3"]

COMPILE_RULES = {
    ".py": Compile._compile_script,
    ".sh": Compile._compile_script,
    ".c": Compile._compile_c,
    ".cc": Compile._compile_cpp,
    ".cpp": Compile._compile_cpp,
    ".pas": Compile._compile_pas,
    ".rs": Compile._compile_rust,
}


def supported_extensions() -> list[str]:
    return list(COMPILE_RULES)
